/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
/* The following description is used by demos/make-demo-list.pl
 * to compile a list of all tests. */
/* DESCRIPTION
 * vibrato demo
 *
 * Function: vibrato_demo
 */


#include <gtk/gtk.h>
#include <gdam/gdambuffer.h>
#include <math.h>
#include <stdio.h>
#include "../gdamchannel.h"

typedef struct _DemoInfo {
	GdamChannel* channel;
	GtkWindow* top_level;
	GtkLabel* filename;
	GtkLabel* hertz_label;
	GtkLabel* beat_label;
	GtkLabel* state;
	GtkAdjustment* beat_adjustment;
	GtkAdjustment* hertz_adjustment;
	GtkAdjustment* min_adjustment;
	GtkAdjustment* range_adjustment;
	guint mixer_id;
	guint buffer_id;
	guint beat_buffer_id;
	guint source_id;
	guint hertz_id;
	guint beat_id;
	guint square_wave, triangle_wave, sine_wave;
	gboolean making_wave_request;
} DemoInfo;

static void indicate(DemoInfo* info, const char* state)
{
	gtk_label_set_text(info->state, state);
}

/*static void stopped_callback(gpointer user_data)
{
	DemoInfo* info = (DemoInfo*)user_data;
	if (info->wave_id == 0 || info->making_wave_request)
		return;
	info->wave_id = 0;
	indicate(info, "ended");
}*/
	
static gdouble get_hertz(DemoInfo* info)
{
	return exp(info->hertz_adjustment->value);
}
static gdouble get_beat(DemoInfo* info)
{
	return exp(info->beat_adjustment->value);
}
static void play_callback(GtkButton* button, DemoInfo* info)
{
	gchar* filename;
	GdamTime time_array[4];
	if (info->source_id != 0 || info->making_wave_request)
		return;
	g_return_if_fail(info->buffer_id != 0);

	info->making_wave_request = TRUE;
	gtk_label_get(info->filename, &filename);
	gdam_debug("create source");
	info->beat_id = gdam_channel_create_source_generic(
				info->channel,
				"GdamWavetable",
				2,
				"buffer_id",
					GDAM_ARG_TYPE_INT,
					info->beat_buffer_id,
				"hertz",
					GDAM_ARG_TYPE_DOUBLE,
					get_beat(info));
	info->hertz_id = gdam_channel_create_source_generic(
				info->channel,
				"GdamWavetable",
				2,
				"buffer_id",
					GDAM_ARG_TYPE_INT,
					info->buffer_id,
				"velocity",
					GDAM_ARG_TYPE_DOUBLE,
					16.0
				//"hertz",
					//GDAM_ARG_TYPE_DOUBLE,
					//get_hertz(info)
					);
	time_array[0] = 0;
	time_array[1] = info->beat_id;
	time_array[2] = 0;
	time_array[3] = info->hertz_id;
	info->source_id = gdam_channel_create_source_generic(
				info->channel,
				"GdamVibrato",
				4,
				"filtered",
				GDAM_ARG_TYPE_INT,
				info->hertz_id,
				"vibrato",
				GDAM_ARG_TYPE_INT,
				info->beat_id,
				"min_rate",
				GDAM_ARG_TYPE_DOUBLE,
				0.5,
				"max_rate",
				GDAM_ARG_TYPE_DOUBLE,
				1.5
				);

	g_assert(info->source_id != 0);
	gdam_debug("gdam_channel_mix_add being called...");
	if (!gdam_channel_mix_add(info->channel,
	                           info->mixer_id,
				   info->source_id,
				   -1))
		g_warning("mix_add failed");
	else
		indicate(info, "playing");
	info->making_wave_request = FALSE;
}

static void stop_callback(GtkButton* button, DemoInfo* info)
{
	if (info->source_id == 0 || info->making_wave_request)
		return;
	if (!gdam_channel_detach_source(info->channel, info->source_id))
		g_warning("detach source failed");
	else
		indicate(info, "stopped");
	info->source_id = 0;
}
static void hertz_changed(GtkAdjustment* adjustment, DemoInfo* info)
{
	gdouble hertz = get_hertz(info);
	char buf[100];
	gdam_channel_source_set_arg(info->channel, info->hertz_id,
	                            "hertz",
				    GDAM_ARG_TYPE_DOUBLE,
				    hertz);
	sprintf(buf, "%.6g", (double)hertz);
	gtk_label_set_text(info->hertz_label, buf);
}
static void beat_changed(GtkAdjustment* adjustment, DemoInfo* info)
{
	gdouble hertz = get_beat(info);
	char buf[100];
	gdam_channel_source_set_arg(info->channel, info->beat_id,
	                            "hertz",
				    GDAM_ARG_TYPE_DOUBLE,
				    hertz);
	sprintf(buf, "%.6g", (double)hertz);
	gtk_label_set_text(info->beat_label, buf);
}
static void on_success(gpointer unused)
{
}
static void on_failure(gpointer unused)
{
	g_error("failed setting arguments");
}
static void adjust_range(GtkAdjustment* adjustment, DemoInfo* info)
{
	GdamArg* arg;
	GdamRequest* request;
	arg = gdam_arg_new_from_valist(2,
	                            "min_rate",
				    GDAM_ARG_TYPE_DOUBLE,
				    info->min_adjustment->value,
	                            "max_rate",
				    GDAM_ARG_TYPE_DOUBLE,
				    info->min_adjustment->value
				    	+ info->range_adjustment->value);
	request = gdam_request_set_args(info->source_id, 2, arg,
					on_success, on_failure, NULL);
	if (!gdam_channel_process_request(info->channel, request))
		g_error("error processing request");
}

#if 0
static void sine_callback(GtkButton* button, DemoInfo* info)
{
	info->buffer_id = info->sine_wave;
}
static void triangle_callback(GtkButton* button, DemoInfo* info)
{
	info->buffer_id = info->triangle_wave;
}
static void square_callback(GtkButton* button, DemoInfo* info)
{
	info->buffer_id = info->square_wave;
}
#endif

/* Helper + macro to cut down repetition in demo1_init(). */
static GtkWidget* _generic_button(const char* label, GtkSignalFunc callback,
                                 gpointer user_data)
{
	GtkWidget* button = gtk_button_new_with_label(label);
	gtk_signal_connect(GTK_OBJECT(button), "clicked", callback, user_data);
	gtk_widget_show(GTK_WIDGET(button));
	return button;
}
#define generic_button(label, callback)				\
		_generic_button(label, GTK_SIGNAL_FUNC(callback), info)

void vibrato_demo(GdamChannel*           channel,
                guint                  mixer_id)
{
	DemoInfo* info = g_new(DemoInfo, 1);
	GtkWidget* hbox;
	GtkWidget* vbox;
	GtkWidget* tmp;
	vbox = gtk_vbox_new(FALSE, 0);
	hbox = gtk_hbox_new(FALSE, 0);

	info->channel = channel;
	info->top_level = (GtkWindow*) gtk_window_new(GTK_WINDOW_TOPLEVEL);
	info->filename = (GtkLabel*) gtk_label_new("???");
	info->state = (GtkLabel*) gtk_label_new("initial state");
	info->hertz_label = (GtkLabel*) gtk_label_new("440");
	info->beat_label = (GtkLabel*) gtk_label_new("1");
	info->making_wave_request = FALSE;
	info->mixer_id = mixer_id;
	info->source_id = 0;
	if ( !gdam_channel_get_server_side_int(channel,
	                                       "wavetables",
	                                       "sine",
					       &info->sine_wave)
	  || !gdam_channel_get_server_side_int(channel,
	                                       "wavetables",
	                                       "square",
					       &info->square_wave)
	  || !gdam_channel_get_server_side_int(channel,
	                                       "wavetables",
	                                       "triangle",
					       &info->triangle_wave)
	  || info->sine_wave == 0
	  || info->square_wave == 0
	  || info->triangle_wave == 0) {
	  	g_warning("error lookup for sine/square/triangle name");
		return;
	}

	g_message("sine=%d, square=%d, triangle=%d",
	       info->sine_wave, info->square_wave, info->triangle_wave);

	info->buffer_id = info->sine_wave;
	info->beat_buffer_id = info->sine_wave;

	gtk_widget_show(GTK_WIDGET(info->filename));
	gtk_box_pack_start_defaults(GTK_BOX(hbox), GTK_WIDGET(info->filename));
	gtk_widget_show(GTK_WIDGET(info->hertz_label));
	gtk_box_pack_start_defaults(GTK_BOX(hbox), GTK_WIDGET(info->hertz_label));
	tmp = gtk_vseparator_new();
	gtk_box_pack_start_defaults(GTK_BOX(hbox), tmp);
	gtk_widget_show(tmp);
	gtk_widget_show(GTK_WIDGET(info->beat_label));
	gtk_box_pack_start_defaults(GTK_BOX(hbox), GTK_WIDGET(info->beat_label));
	gtk_widget_show(GTK_WIDGET(info->state));
	gtk_widget_show(hbox);
	gtk_box_pack_start_defaults(GTK_BOX(hbox), GTK_WIDGET(info->state));

	gtk_box_pack_start_defaults(GTK_BOX(vbox), GTK_WIDGET(hbox));

	hbox = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                        generic_button("Play", play_callback));
	gtk_box_pack_start_defaults(GTK_BOX(hbox),
	                        generic_button("Stop", stop_callback));
	gtk_widget_show(hbox);
	gtk_box_pack_start_defaults(GTK_BOX(vbox), GTK_WIDGET(hbox));

	info->hertz_adjustment = (GtkAdjustment*)gtk_adjustment_new(
						log(440), log(20), log(30000),
						1, 1, 1);
	info->beat_adjustment = (GtkAdjustment*)gtk_adjustment_new(
						log(1), log(0.1), log(1000.0),
						1, 1, 1);
	info->min_adjustment = GTK_ADJUSTMENT(gtk_adjustment_new(
						1.0, 0.1, 5.0,
						1, 1, 1));
	info->range_adjustment = GTK_ADJUSTMENT(gtk_adjustment_new(
						0.0, 0.0, 5.0,
						1, 1, 1));
	gtk_signal_connect(GTK_OBJECT(info->hertz_adjustment),
	                   "value_changed",
			   GTK_SIGNAL_FUNC(hertz_changed),
			   info);
	gtk_signal_connect(GTK_OBJECT(info->beat_adjustment),
	                   "value_changed",
			   GTK_SIGNAL_FUNC(beat_changed),
			   info);
	gtk_signal_connect(GTK_OBJECT(info->min_adjustment),
	                   "value_changed",
			   GTK_SIGNAL_FUNC(adjust_range),
			   info);
	gtk_signal_connect(GTK_OBJECT(info->range_adjustment),
	                   "value_changed",
			   GTK_SIGNAL_FUNC(adjust_range),
			   info);
	{
		GtkWidget* bar = gtk_hscrollbar_new(info->hertz_adjustment);
		gtk_widget_show(bar);
		gtk_box_pack_start_defaults(GTK_BOX(vbox), bar);

		bar = gtk_hscrollbar_new(info->beat_adjustment);
		gtk_widget_show(bar);
		gtk_box_pack_start_defaults(GTK_BOX(vbox), bar);

		bar = gtk_hscrollbar_new(info->min_adjustment);
		gtk_widget_show(bar);
		gtk_box_pack_start_defaults(GTK_BOX(vbox), bar);

		bar = gtk_hscrollbar_new(info->range_adjustment);
		gtk_widget_show(bar);
		gtk_box_pack_start_defaults(GTK_BOX(vbox), bar);
	}


	gtk_widget_show(vbox);
	gtk_container_add(GTK_CONTAINER(info->top_level), GTK_WIDGET(vbox));
	gtk_widget_show(GTK_WIDGET(info->top_level));
}
