/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#ifndef __GDAM_SOUND_SPEC_H_
#define __GDAM_SOUND_SPEC_H_

#include <gdam/gdambuffer.h>
#include <gdam/gdamprinter.h>

typedef struct _GdamSoundSpecAny GdamSoundSpecAny;
typedef union _GdamSoundSpec GdamSoundSpec;
typedef struct _GdamSoundSpecMp3 GdamSoundSpecMp3;
typedef struct _GdamSoundSpecMixElement GdamSoundSpecMixElement;
typedef struct _GdamSoundSpecMix GdamSoundSpecMix;
typedef struct _GdamSoundSpecModulator GdamSoundSpecModulator;
typedef struct _GdamSoundSpecEchoElement GdamSoundSpecEchoElement;
typedef struct _GdamSoundSpecEcho GdamSoundSpecEcho;
typedef struct _GdamSoundSpecGain GdamSoundSpecGain;
typedef struct _GdamSoundSpecResample GdamSoundSpecResample;
typedef struct _GdamSoundSpecById GdamSoundSpecById;
typedef struct _GdamSoundSpecCache GdamSoundSpecCache;
typedef struct _GdamSoundSpecSequencerElement GdamSoundSpecSequencerElement;
typedef struct _GdamSoundSpecSequencer GdamSoundSpecSequencer;
typedef struct _GdamSoundSpecWav GdamSoundSpecWav;
typedef struct _GdamSoundSpecWavetable GdamSoundSpecWavetable;
typedef struct _GdamSoundSpecWalker GdamSoundSpecWalker;

struct _GdamSoundSpecAny {
	int			type;
	void		      (*destroy)     (GdamSoundSpec*  spec);
	void                  (*to_buffer)   (GdamSoundSpec*  spec,
	                                      GdamBuffer*     into);
	void		      (*pretty_print)(GdamSoundSpec*  spec,
	                                      GdamPrinter*    printer);
};

struct _GdamSoundSpecMp3 {
	GdamSoundSpecAny	base;
	char*			mp3_filename;
	GdamTime		start;
	GdamTime		length;
};

struct _GdamSoundSpecMixElement {
	GdamTime		start_time;
	GdamSoundSpec*		spec;
};

struct _GdamSoundSpecMix {
	GdamSoundSpecAny	base;
	int			terminates;
	GList*			first;   /* of GdamSoundSpecMixElement */
	GList*			last;   /* of GdamSoundSpecMixElement */
};

struct _GdamSoundSpecModulator {
	GdamSoundSpecAny	base;
	int			terminates;
	GList*			first;   /* of GdamSoundSpecMixElement */
	GList*			last;   /* of GdamSoundSpecMixElement */
};

struct _GdamSoundSpecGain {
	GdamSoundSpecAny	base;
	int			level;
	GdamSoundSpec*		subsource;
};

struct _GdamSoundSpecResample {
	GdamSoundSpecAny	base;
	int			rate;
	GdamSoundSpec*		subsource;
};

struct _GdamSoundSpecEchoElement {
	int			volume;
	GdamTime		delay;
};

struct _GdamSoundSpecEcho {
	GdamSoundSpecAny	base;
	gboolean                reverb;
	int			num_echoes;
	GdamSoundSpecEchoElement* echoes;
	GdamSoundSpec*		subsource;
};

struct _GdamSoundSpecCache {
	GdamSoundSpecAny	base;
	guint                   buffer_id;
	gboolean                was_loop;
	int			start_sample;
};

struct _GdamSoundSpecWavetable {
	GdamSoundSpecAny	base;
	guint                   buffer_id;
	gint64			hertz_32;
};

struct _GdamSoundSpecWalker {
	GdamSoundSpecAny	base;
	guint                   buffer_id;
};

struct _GdamSoundSpecSequencerElement {
	int			buffer_id;
	int			start_offset;
	int			resample_b16;
	int			gain_b16;
	int			sample_offset;
	int			output_length;
};

struct _GdamSoundSpecSequencer {
	GdamSoundSpecAny	base;
	int			total_time;
	int			num_elements;
	GdamSoundSpecSequencerElement* elements;
};

struct _GdamSoundSpecWav {
	GdamSoundSpecAny	base;
	char*			wav_filename;
};

struct _GdamSoundSpecById {
	GdamSoundSpecAny	base;
	int			source_id;
};

union _GdamSoundSpec {
	int			type;
	GdamSoundSpecAny	any_spec;
	GdamSoundSpecMix	mix_spec;
	GdamSoundSpecModulator	modulator_spec;
	GdamSoundSpecMp3	mp3_spec;
	GdamSoundSpecGain	gain_spec;
	GdamSoundSpecResample	resample_spec;
	GdamSoundSpecEcho	echo_spec;
	GdamSoundSpecCache	cache_spec;
	GdamSoundSpecSequencer	sequencer_spec;
	GdamSoundSpecWavetable	wavetable_spec;
	GdamSoundSpecWav	wav_spec;
	GdamSoundSpecWalker	walker_spec;
	GdamSoundSpecById	by_id_spec;
};

GdamSoundSpec* gdam_sound_spec_parse  (GdamBuffer*    buffer);

void           gdam_sound_spec_pretty_print(GdamSoundSpec* spec,
                                       GdamPrinter*   printer);

/* In the next function the sound specs and the list that contains them
 * are freed by the mix. */
GdamSoundSpec* gdam_sound_spec_mix_list(int           terminates,
				       GList*	      elements);

GdamSoundSpec* gdam_sound_spec_mix    (int            terminates,
                                       int            num_subs,
				       ...);
GdamSoundSpec* gdam_sound_spec_modulator
                                      (int          terminates,
                                       int            num_subs,
				       ...);
GdamSoundSpec* gdam_sound_spec_modulator_list
                                      (int           terminates,
				       GList*	      elements);
GdamSoundSpec* gdam_sound_spec_mp3    (const char*    filename,
                                       GdamTime       start_time,
				       GdamTime       length);
GdamSoundSpec* gdam_sound_spec_gain   (int            level,
                                       GdamSoundSpec* sub);
GdamSoundSpec* gdam_sound_spec_resample(int           rate,
                                       GdamSoundSpec* sub);
GdamSoundSpec* gdam_sound_spec_echo   (GdamSoundSpec* sub,
                                       gboolean       reverb,
                                       int            num_echoes,
				       ...);	/* volume/delay pairs... */
GdamSoundSpec* gdam_sound_spec_echo_v (GdamSoundSpec* sub,
                                       gboolean       reverb,
                                       int            num_echoes,
	                               const GdamSoundSpecEchoElement* echoes);
GdamSoundSpec* gdam_sound_spec_cache  (int            buffer_id,
                                       gboolean       was_loop,
				       int            start_sample);
GdamSoundSpec* gdam_sound_spec_sequencer
                                      (int            length,
                                       int            num_subs,
                 const GdamSoundSpecSequencerElement* element);

GdamSoundSpec* gdam_sound_spec_walker (int            buffer_id);
				       
GdamSoundSpec* gdam_sound_spec_wav    (const char*    filename);
GdamSoundSpec* gdam_sound_spec_wavetable(int          buffer_id,
				       gint64	      hertz_32);
GdamSoundSpec* gdam_sound_spec_by_id  (int            source_id);

void           gdam_sound_spec_write  (GdamSoundSpec* spec,
                                       GdamBuffer*    output_buffer);
void           gdam_sound_spec_destroy(GdamSoundSpec* spec);

#endif
